# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
from __future__ import annotations

project = "jupyter_events"
copyright = "2019, Project Jupyter"
author = "Project Jupyter"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions: list = ["myst_parser", "jupyterlite_sphinx"]

try:
    import enchant  # noqa: F401

    extensions += ["sphinxcontrib.spelling"]
except ImportError:
    pass

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffix = [".rst", ".md"]


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "_contents",
    "Thumbs.db",
    ".DS_Store",
    "demo/demo-notebook.ipynb",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/jupyter_logo.png"

master_doc = "index"

# Configure jupyterlite to import jupyter_events package
jupyterlite_contents = ["demo/demo-notebook.ipynb"]

html_theme_options = {
    "logo": {
        "text": "Jupyter Events",
    },
    "navigation_with_keys": False,
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/jupyter/jupyter_events",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fab fa-github-square",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        },
        {
            "name": "jupyter.org",
            "url": "https://jupyter.org",
            "icon": "_static/jupyter_logo.png",
            "type": "local",
        },
    ],
}



from __future__ import annotations

import os
from pathlib import Path

import pytest
from jsonschema.exceptions import ValidationError

from jupyter_events import yaml
from jupyter_events.schema import (
    EventSchema,
    EventSchemaFileAbsent,
    EventSchemaLoadingError,
    EventSchemaUnrecognized,
)
from jupyter_events.validators import validate_schema

from .utils import SCHEMA_PATH

BAD_SCHEMAS = [
    ["reserved-property.yaml", "Properties starting with 'dunder'"],
    ["nested-reserved-property.yaml", "Properties starting with 'dunder'"],
    ["bad-id.yaml", "'not-a-uri' is not a 'uri'"],
]

GOOD_SCHEMAS = ["array.yaml", "nested-array.yaml", "basic.yaml"]


@pytest.mark.parametrize("schema_file,validation_error_msg", BAD_SCHEMAS)
def test_bad_validations(schema_file, validation_error_msg):
    """
    Validation fails because the schema is missing
    a redactionPolicies field.
    """
    # Read the schema file
    with Path.open(SCHEMA_PATH / "bad" / schema_file) as f:
        schema = yaml.loads(f)
    # Assert that the schema files for a known reason.
    with pytest.raises(ValidationError) as err:
        validate_schema(schema)
    assert validation_error_msg in err.value.message


def test_file_absent():
    """Validation fails because file does not exist at path."""
    with pytest.raises(EventSchemaFileAbsent):
        EventSchema(Path("asdf.txt"))


def test_string_intended_as_path():
    """Ensure EventSchema returns a helpful error message if user passes a
    string intended as a Path."""
    expected_msg_contents = "Paths to schema files must be explicitly wrapped in a Pathlib object."
    str_path = os.path.join(SCHEMA_PATH, "good", "some_schema.yaml")  # noqa: PTH118
    with pytest.raises(EventSchemaLoadingError) as e:
        EventSchema(str_path)

    assert expected_msg_contents in str(e)


def test_unrecognized_type():
    """Validation fails because file is not of valid type."""
    with pytest.raises(EventSchemaUnrecognized):
        EventSchema(9001)  # type:ignore[arg-type]


def test_invalid_yaml():
    """Validation fails because deserialized schema is not a dictionary."""
    path = SCHEMA_PATH / "bad" / "invalid.yaml"
    with pytest.raises(EventSchemaLoadingError):
        EventSchema(path)


def test_valid_json():
    """Ensure EventSchema accepts JSON files."""
    path = SCHEMA_PATH / "good" / "basic.json"
    EventSchema(path)


@pytest.mark.parametrize("schema_file", GOOD_SCHEMAS)
def test_good_validations(schema_file):
    """Ensure validation passes for good schemas."""
    # Read the schema file
    with Path.open(SCHEMA_PATH / "good" / schema_file) as f:
        schema = yaml.loads(f)
    # assert that no exception gets raised
    validate_schema(schema)


@pytest.mark.parametrize(
    "schema",
    [
        # Non existent paths
        "non-existent-file.yml",
        "non/existent/path",
        "non/existent/path/file.yaml",
        # Valid yaml string, but not a valid object
        "random string",
    ],
)
def test_loading_string_error(schema):
    with pytest.raises(EventSchemaLoadingError):
        EventSchema(schema)



from __future__ import annotations

import io
import logging

import pytest

from jupyter_events.logger import EventLogger, SchemaNotRegistered
from jupyter_events.schema import EventSchema

from .utils import SCHEMA_PATH


@pytest.fixture
def schema():
    # Read schema from path.
    schema_path = SCHEMA_PATH / "good" / "basic.yaml"
    return EventSchema(schema=schema_path)


@pytest.fixture
def jp_event_schemas(schema):
    return [schema]


async def test_listener_function(jp_event_logger, schema):
    event_logger = jp_event_logger
    listener_was_called = False

    async def my_listener(logger: EventLogger, schema_id: str, data: dict) -> None:
        nonlocal listener_was_called
        listener_was_called = True

    # Add the modifier
    event_logger.add_listener(schema_id=schema.id, listener=my_listener)
    event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})
    await event_logger.gather_listeners()
    assert listener_was_called
    # Check that the active listeners are cleaned up.
    assert len(event_logger._active_listeners) == 0


async def test_listener_function_str_annotations(jp_event_logger, schema):
    event_logger = jp_event_logger
    listener_was_called = False

    async def my_listener(logger: EventLogger, schema_id: str, data: dict) -> None:
        nonlocal listener_was_called
        listener_was_called = True

    # Add the modifier
    event_logger.add_listener(schema_id=schema.id, listener=my_listener)
    event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})
    await event_logger.gather_listeners()
    assert listener_was_called
    # Check that the active listeners are cleaned up.
    assert len(event_logger._active_listeners) == 0


async def test_remove_listener_function(jp_event_logger, schema):
    event_logger = jp_event_logger
    listener_was_called = False

    async def my_listener(logger: EventLogger, schema_id: str, data: dict) -> None:
        nonlocal listener_was_called
        listener_was_called = True

    # Add the modifier
    event_logger.add_listener(schema_id=schema.id, listener=my_listener)
    event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})
    await event_logger.gather_listeners()
    assert listener_was_called

    # Check that the active listeners are cleaned up.
    assert len(event_logger._active_listeners) == 0

    event_logger.remove_listener(listener=my_listener)
    assert len(event_logger._modified_listeners[schema.id]) == 0
    assert len(event_logger._unmodified_listeners[schema.id]) == 0


async def test_listener_that_raises_exception(jp_event_logger, schema):
    event_logger = jp_event_logger

    # Get an application logger that will show the exception
    app_log = event_logger.log
    log_stream = io.StringIO()
    h = logging.StreamHandler(log_stream)
    app_log.addHandler(h)

    async def listener_raise_exception(logger: EventLogger, schema_id: str, data: dict) -> None:
        raise Exception("This failed")  # noqa

    event_logger.add_listener(schema_id=schema.id, listener=listener_raise_exception)
    event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})

    await event_logger.gather_listeners()

    # Check that the exception was printed to the logs
    h.flush()
    log_output = log_stream.getvalue()
    assert "This failed" in log_output
    # Check that the active listeners are cleaned up.
    assert len(event_logger._active_listeners) == 0


async def test_bad_listener_does_not_break_good_listener(jp_event_logger, schema):
    event_logger = jp_event_logger

    # Get an application logger that will show the exception
    app_log = event_logger.log
    log_stream = io.StringIO()
    h = logging.StreamHandler(log_stream)
    app_log.addHandler(h)

    listener_was_called = False

    async def listener_raise_exception(logger: EventLogger, schema_id: str, data: dict) -> None:
        raise Exception("This failed")  # noqa

    async def my_listener(logger: EventLogger, schema_id: str, data: dict) -> None:
        nonlocal listener_was_called
        listener_was_called = True

    # Add a bad listener and a good listener and ensure that
    # emitting still works and the bad listener's exception is is logged.
    event_logger.add_listener(schema_id=schema.id, listener=listener_raise_exception)
    event_logger.add_listener(schema_id=schema.id, listener=my_listener)

    event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})

    await event_logger.gather_listeners()

    # Check that the exception was printed to the logs
    h.flush()
    log_output = log_stream.getvalue()
    assert "This failed" in log_output
    assert listener_was_called
    # Check that the active listeners are cleaned up.
    assert len(event_logger._active_listeners) == 0


@pytest.mark.parametrize(
    # Make sure no schemas are added at the start of this test.
    "jp_event_schemas",
    [
        # Empty events list.
        []
    ],
)
async def test_listener_added_before_schemas_passes(jp_event_logger, schema):
    # Ensure there are no schemas listed.
    assert len(jp_event_logger.schemas.schema_ids) == 0

    listener_was_called = False

    async def my_listener(logger: EventLogger, schema_id: str, data: dict) -> None:
        nonlocal listener_was_called
        listener_was_called = True

    # Add the listener without any schemas
    jp_event_logger.add_listener(schema_id=schema.id, listener=my_listener)

    # Proof that emitting the event won't success
    with pytest.warns(SchemaNotRegistered):
        jp_event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})

    assert not listener_was_called

    # Now register the event and emit.
    jp_event_logger.register_event_schema(schema)

    # Try emitting the event again and ensure the listener saw it.
    jp_event_logger.emit(schema_id=schema.id, data={"prop": "hello, world"})
    await jp_event_logger.gather_listeners()
    assert listener_was_called
    # Check that the active listeners are cleaned up.
    assert len(jp_event_logger._active_listeners) == 0



"""Event validators."""
from __future__ import annotations

import pathlib
import warnings
from typing import Any

import jsonschema
from jsonschema import Draft7Validator, ValidationError
from referencing import Registry
from referencing.jsonschema import DRAFT7

from . import yaml
from .utils import JupyterEventsVersionWarning

draft7_format_checker = (
    Draft7Validator.FORMAT_CHECKER
    if hasattr(Draft7Validator, "FORMAT_CHECKER")
    else jsonschema.draft7_format_checker
)


METASCHEMA_PATH = pathlib.Path(__file__).parent.joinpath("schemas")

EVENT_METASCHEMA_FILEPATH = METASCHEMA_PATH.joinpath("event-metaschema.yml")
EVENT_METASCHEMA = yaml.load(EVENT_METASCHEMA_FILEPATH)

EVENT_CORE_SCHEMA_FILEPATH = METASCHEMA_PATH.joinpath("event-core-schema.yml")
EVENT_CORE_SCHEMA = yaml.load(EVENT_CORE_SCHEMA_FILEPATH)

PROPERTY_METASCHEMA_FILEPATH = METASCHEMA_PATH.joinpath("property-metaschema.yml")
PROPERTY_METASCHEMA = yaml.load(PROPERTY_METASCHEMA_FILEPATH)

SCHEMA_STORE = {
    EVENT_METASCHEMA["$id"]: EVENT_METASCHEMA,
    PROPERTY_METASCHEMA["$id"]: PROPERTY_METASCHEMA,
    EVENT_CORE_SCHEMA["$id"]: EVENT_CORE_SCHEMA,
}

resources = [
    DRAFT7.create_resource(each)
    for each in (EVENT_METASCHEMA, PROPERTY_METASCHEMA, EVENT_CORE_SCHEMA)
]
METASCHEMA_REGISTRY: Registry[Any] = resources @ Registry()

JUPYTER_EVENTS_SCHEMA_VALIDATOR = Draft7Validator(
    schema=EVENT_METASCHEMA,
    registry=METASCHEMA_REGISTRY,
    format_checker=draft7_format_checker,
)

JUPYTER_EVENTS_CORE_VALIDATOR = Draft7Validator(
    schema=EVENT_CORE_SCHEMA,
    registry=METASCHEMA_REGISTRY,
    format_checker=draft7_format_checker,
)


def validate_schema(schema: dict[str, Any]) -> None:
    """Validate a schema dict."""
    try:
        # If the `version` attribute is an integer, coerce to string.
        # TODO: remove this in a future version.
        if "version" in schema and isinstance(schema["version"], int):
            schema["version"] = str(schema["version"])
            msg = (
                "The `version` property of an event schema must be a string. "
                "It has been type coerced, but in a future version of this "
                "library, it will fail to validate. Please update schema: "
                f"{schema['$id']}"
            )
            warnings.warn(JupyterEventsVersionWarning(msg), stacklevel=2)
        # Validate the schema against Jupyter Events metaschema.
        JUPYTER_EVENTS_SCHEMA_VALIDATOR.validate(schema)
    except ValidationError as err:
        reserved_property_msg = " does not match '^(?!__.*)'"
        if reserved_property_msg in str(err):
            idx = str(err).find(reserved_property_msg)
            bad_property = str(err)[:idx].strip()
            msg = (
                f"{bad_property} is an invalid property name because it "
                "starts with `__`. Properties starting with 'dunder' "
                "are reserved as special meta-fields for Jupyter Events to use."
            )
            raise ValidationError(msg) from err
        raise err
